#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 25/2/24 21:54 PM
# @Author  : zhoutao3
# @File    : dnn_entry.py
# @Email   : zhoutao3@genomics.cn
'''
import os.path as osp
import pandas as pd
import scanpy as sc # type: ignore
import numpy as np
import scipy.sparse as sp
import random
import torch
import torch_geometric # type: ignore
from matplotlib import pyplot as plt

from .cell_type_ann_model import SpatialModelTrainer # type: ignore

random.seed(0)
np.random.seed(0)
torch.manual_seed(0)
torch.cuda.manual_seed(0)
torch.cuda.manual_seed_all(0)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True


def load_data(data_path, filter_mt=True, min_cells=10, min_counts=300, max_percent=98.0):
    """
    loading and processing dataset
    :param data_path: Note that, the input data must be raw, can not do any preprocessing!
    :param filter_mt: Whether to filter MT- genes. default, True
    :param min_cells: Whether to filter genes. default, 10
    :param min_counts: Whether to filter cells. default, 300
    :param max_percent: Whether to filter cells, Range: (0, 100). default, 98.0
    :return:
    """
    print("======> Loading data...")
    adata = sc.read_h5ad(data_path)
    print('  Original Data Info: %d cells × %d genes.' % (adata.shape[0], adata.shape[1]))

    if filter_mt:
        adata.var["mt"] = adata.var_names.str.startswith(("MT-", "Mt-", "mt-"))
        sc.pp.calculate_qc_metrics(adata, qc_vars=["mt"], inplace=True)
        adata = adata[adata.obs["pct_counts_mt"] < 10].copy() # type: ignore
    if min_counts > 0:
        sc.pp.filter_cells(adata, min_counts=min_counts)
    if min_cells > 0:
        sc.pp.filter_genes(adata, min_cells=min_cells)
    if max_percent < 100:
        max_counts = np.percentile(adata.X.sum(1).reshape(-1).tolist()[0], max_percent) # type: ignore
        sc.pp.filter_cells(adata, max_counts=max_counts)

    # adata_X_sparse_backup = adata.X.copy()
    print('  After Preprocessing Data Info: %d cells × %d genes.' % (adata.shape[0], adata.shape[1]))
    if sp.issparse(adata.X):
        adata.X = adata.X.toarray() # type: ignore
    return adata


def transfer_from_sc_data(adata, model_choice, model_path, gpu="0"):
    """
    :param adata:
    :param dnn_path: Pre-trained DNN model save path
    :param gpu: gpu number
    :return:
    """
    print("========> Transfering from sc-dataset...")
    if gpu is not None and torch.cuda.is_available():
        device = torch.device("cuda:{}".format(gpu))
    else:
        device = torch.device("cpu")
    if model_choice == "dnn":
        checkpoint = torch.load(model_path)
        dnn_model = checkpoint["model"].to(device)
        dnn_model.eval()

        marker_genes = checkpoint["marker_genes"]
        gene_indices = adata.var_names.get_indexer(marker_genes)
        adata_X = np.pad(adata.X, ((0, 0), (0, 1)))[:, gene_indices].copy()
        norm_factor = np.linalg.norm(adata_X, axis=1, keepdims=True)
        norm_factor[norm_factor == 0] = 1
        dnn_inputs = torch.Tensor(adata_X / norm_factor).split(checkpoint["batch_size"])
        # Inference with DNN model.
        dnn_predictions = []
        with torch.no_grad():
            for batch_idx, inputs in enumerate(dnn_inputs):
                inputs = inputs.to(device)
                outputs = dnn_model(inputs)
                dnn_predictions.append(outputs.detach().cpu().numpy())
        label_names = checkpoint['label_names']
        adata.obsm['psuedo_label'] = np.concatenate(dnn_predictions)
        adata.obs['psuedo_class'] = pd.Categorical([label_names[i] for i in adata.obsm['psuedo_label'].argmax(1)])
        adata.uns['psuedo_classes'] = label_names
        return adata
    elif model_choice == "xgboost":
        import pickle
        with open(model_path, 'rb') as f:
            model, dic, marker_genes = pickle.load(f)
        gene_indices = adata.var_names.get_indexer(marker_genes)
        adata_X = np.pad(adata.X, ((0, 0), (0, 1)))[:, gene_indices].copy()
        xgboost_predictions = model.predict_proba(adata_X)
        adata.obsm['psuedo_label'] = xgboost_predictions
        adata.obs['psuedo_class'] = pd.Categorical([dic[i] for i in adata.obsm['psuedo_label'].argmax(1)])
        adata.uns['psuedo_classes'] = dic
        return adata
    return adata


def distribution_fine_tune(adata, pca_dim=200, k_graph=30, edge_weight=True, epochs=200, w_cls=20, w_dae=1., w_gae=1.,
                           gpu="0", save_path="./output"):
    """
    :param adata:
    :param pca_dim: PCA dims, default=200
    :param k_graph: neighbors number, default=30
    :param edge_weight: Add edge weight to the graph model, default=True
    :param epochs: GCN training epochs, default=200
    :param w_cls: class num weight, default=20
    :param w_dae: dnn weight
    :param w_gae: gcn weight
    :param gpu: gpu number
    :param save_path: results save path
    :return:
    """
    if gpu is not None and torch.cuda.is_available():
        device = torch.device("cuda:{}".format(gpu))
    else:
        device = torch.device("cpu")

    print("========> Model Training...")
    sc.pp.normalize_per_cell(adata, counts_per_cell_after=1e4)
    sc.pp.log1p(adata)
    adata.X = (adata.X - adata.X.mean(0)) / adata.X.std(0)
    gene_mat = torch.Tensor(adata.X)
    if pca_dim:
        u, s, v = torch.pca_lowrank(gene_mat, pca_dim)
        gene_mat = torch.matmul(gene_mat, v)
    # u, s, v = torch.pca_lowrank(gene_mat, pca_dim)
    # gene_mat = torch.matmul(gene_mat, v)
    cell_coo = torch.Tensor(adata.obsm['spatial'])
    data = torch_geometric.data.Data(x=gene_mat, pos=cell_coo) # type: ignore
    data = torch_geometric.transforms.KNNGraph(k=k_graph, loop=True)(data) # type: ignore
    data.y = torch.Tensor(adata.obsm['psuedo_label'])

    # Make distances as edge weights.
    if edge_weight:
        data = torch_geometric.transforms.Distance()(data) # type: ignore
        data.edge_weight = 1 - data.edge_attr[:, 0]
    else:
        data.edge_weight = torch.ones(data.edge_index.size(1))

    # Train self-supervision model.
    input_dim = data.num_features
    num_classes = len(adata.uns['psuedo_classes'])
    trainer = SpatialModelTrainer(input_dim, num_classes, device=device)
    trainer.train(data, epochs, w_cls, w_dae, w_gae)
    trainer.save_checkpoint(osp.join(save_path, "model.bgi"))

    # Inference.
    print('\n==> Inferencing...')
    predictions = trainer.valid(data)
    celltype_pred = pd.Categorical([adata.uns['psuedo_classes'][i] for i in predictions])

    # Save results.
    result = pd.DataFrame({'cell': adata.obs_names.tolist(), 'celltype_pred': celltype_pred})
    result.to_csv(osp.join(save_path, "model.csv"), index=False)
    adata.obs['celltype_pred'] = pd.Categorical(celltype_pred) # type: ignore
    # adata.X = adata_X_sparse_backup

    # --------------------------------------------------
    adata.obsm["X_pca"] = gene_mat.detach().cpu().numpy()
    adata.uns = None
    adata.write(osp.join(save_path, "adata.h5ad"))

    # Save visualization.
    spot_size = 30
    psuedo_top100 = adata.obs['psuedo_class'].to_numpy()
    other_classes = list(pd.value_counts(adata.obs['psuedo_class'])[100:].index)
    psuedo_top100[adata.obs['psuedo_class'].isin(other_classes)] = 'Others'
    adata.obs['psuedo_top100'] = pd.Categorical(psuedo_top100)
    sc.pl.spatial(adata, img_key=None, color=['psuedo_top100'], spot_size=spot_size, show=False)
    plt.savefig(osp.join(save_path, "psuedo_top100.pdf"), bbox_inches='tight', dpi=150)
    sc.pl.spatial(adata, img_key=None, color=['celltype_pred'], spot_size=spot_size, show=False)
    plt.savefig(osp.join(save_path, "celltype_pred.pdf"), bbox_inches='tight', dpi=150)
    print("Done!")
'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2/23/24 10:29 AM
# @Author  : zhoutao3
# @File    : cell_type_ann_model.py
# @Software: VSCode
# @Email   : zhoutao3@genomics.cn

import time
import warnings
import torch.nn.functional as F
import torch
import torch.nn as nn
from torch_geometric.nn import GCNConv, VGAE, GATConv,GATv2Conv,GraphConv,GINConv, SAGEConv # type: ignore
from torch_geometric.utils import remove_self_loops, add_self_loops, negative_sampling # type: ignore
import xgboost as xgb # type: ignore
from xgboost import XGBClassifier # type: ignore
warnings.filterwarnings("ignore")

'''
PyTorch Geometric (PyG) 提供了许多图神经网络层，其中一些可以直接替换 GCNConv 而不需要更改参数。以下是一些例子：

ChebConv: ChebNet 卷积层，需要额外的参数 K，表示使用的切比雪夫多项式的阶数。

SAGEConv: GraphSAGE 卷积层。

GATConv: 图注意力网络 (GAT) 卷积层。发现很容易 聚合 临近节点的信息

GINConv: 图同构网络 (GIN) 卷积层。无法正确调用

GraphConv: 图卷积网络层，这是一个更一般的图卷积操作。发现 Psuedo-Acc 非常的低

ARMAConv: 自回归移动平均 (ARMA) 卷积层。

'''

class GraphEncoder(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels):
        super(GraphEncoder, self).__init__()
        # self.gc_feat = GCNConv(in_channels, hidden_channels)
        # self.gc_mean = GCNConv(hidden_channels, out_channels)
        # self.gc_logstd = GCNConv(hidden_channels, out_channels)
        # self.gc_feat = GATConv(in_channels, hidden_channels)
        # self.gc_mean = GATConv(hidden_channels, out_channels)
        # self.gc_logstd = GATConv(hidden_channels, out_channels)
        self.gc_feat = GCNConv(in_channels, hidden_channels,)
        self.gc_mean = GATConv(hidden_channels, out_channels)
        self.gc_logstd = GCNConv(hidden_channels, out_channels)
        
    def forward(self, x, edge_index, edge_weight):
        x = self.gc_feat(x, edge_index, edge_weight).relu()
        mean = self.gc_mean(x, edge_index, edge_weight)
        logstd = self.gc_logstd(x, edge_index, edge_weight)
        return mean, logstd
'''
from torch.nn import Sequential, Linear, ReLU

class GraphEncoder(nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels):
        super(GraphEncoder, self).__init__()

        nn1 = Sequential(Linear(in_channels, hidden_channels), ReLU(), Linear(hidden_channels, hidden_channels))
        self.gc_feat = GINConv(nn1)
        # x = self.gc_feat(x, edge_index)

        nn2 = Sequential(Linear(hidden_channels, out_channels), ReLU(), Linear(out_channels, out_channels))
        self.gc_mean = GINConv(nn2)
        # mean = self.gc_mean(x, edge_index)

        nn3 = Sequential(Linear(hidden_channels, out_channels), ReLU(), Linear(out_channels, out_channels))
        self.gc_logstd = GINConv(nn3)
        # logstd = self.gc_logstd(x, edge_index)

    def forward(self, x, edge_index):
        x = self.gc_feat(x, edge_index).relu()
        mean = self.gc_mean(x, edge_index)
        logstd = self.gc_logstd(x, edge_index)
        return mean, logstd
'''

def full_block(in_features, out_features, drop_rate=0.2):
    return nn.Sequential(
        nn.Linear(in_features, out_features),
        nn.BatchNorm1d(out_features, momentum=0.01, eps=0.001),
        nn.ELU(),
        nn.Dropout(p=drop_rate)
    )


class KDLoss(nn.Module):
    def __init__(self, T):
        super(KDLoss, self).__init__()
        self.T = T

    def forward(self, input, target):
        return nn.KLDivLoss()(
            (input / self.T).log_softmax(1),
            (target / self.T).softmax(1)
        ) * self.T * self.T


class SpatialModel(nn.Module):
    def __init__(self, input_dim, num_classes, gae_dim, dae_dim, feat_dim):
        super(SpatialModel, self).__init__()
        self.input_dim = input_dim
        self.num_classes = num_classes
        self.gae_dim = gae_dim
        self.dae_dim = dae_dim
        self.feat_dim = feat_dim
        self.fcat_dim = self.dae_dim[1] + self.gae_dim[1]
        self.encoder = nn.Sequential(full_block(self.input_dim, self.dae_dim[0]),
                                     full_block(self.dae_dim[0], self.dae_dim[1]))
        self.decoder = nn.Linear(self.feat_dim, self.input_dim)
        self.vgae = VGAE(GraphEncoder(self.dae_dim[1], self.gae_dim[0], self.gae_dim[1]))
        self.feat_fc_x = nn.Sequential(nn.Linear(self.fcat_dim, self.feat_dim), nn.ELU())
        self.feat_fc_g = nn.Sequential(nn.Linear(self.fcat_dim, self.feat_dim), nn.ELU())
        self.classifier = nn.Linear(self.fcat_dim, self.num_classes)

    def forward(self, x, edge_index, edge_weight):
        feat_x = self.encoder(x)
        feat_g = self.vgae.encode(feat_x, edge_index, edge_weight)
        feat = torch.cat([feat_x, feat_g], 1)
        feat_x = self.feat_fc_x(feat)
        feat_g = self.feat_fc_g(feat)
        x_dec = self.decoder(feat_x)
        dae_loss = F.mse_loss(x_dec, x)
        gae_loss = self.recon_loss(feat_g, edge_weight, edge_index) + 1 / len(x) * self.vgae.kl_loss()
        cls = self.classifier(feat)
        return cls, dae_loss, gae_loss

    def recon_loss(self, z, edge_weight, pos_edge_index, neg_edge_index=None):
        pos_dec = self.vgae.decoder(z, pos_edge_index, sigmoid=False)
        pos_loss = F.binary_cross_entropy_with_logits(pos_dec, edge_weight)
        pos_edge_index, _ = remove_self_loops(pos_edge_index)
        pos_edge_index, _ = add_self_loops(pos_edge_index)
        if neg_edge_index is None:
            neg_edge_index = negative_sampling(pos_edge_index, z.size(0))
        neg_dec = self.vgae.decoder(z, neg_edge_index, sigmoid=False)
        neg_loss = -F.logsigmoid(-neg_dec).mean()
        return pos_loss + neg_loss


class SpatialModelTrainer:
    def __init__(self, input_dim, num_classes, device):
        self.scaler = None
        self.scheduler = None
        self.optimizer = None
        self.criterion = None
        self.model = None
        self.device = device
        self.set_model(input_dim, num_classes)
        self.set_optimizer()

    def set_model(self, input_dim, num_classes):
        gae_dim, dae_dim, feat_dim = [32, 8], [100, 20], 64
        self.model = SpatialModel(input_dim, num_classes, gae_dim, dae_dim, feat_dim).to(self.device)
        self.criterion = KDLoss(1)

    def set_optimizer(self):
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=0.01, weight_decay=0.0001) # type: ignore
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, 1, gamma=1.0)
        self.scaler = torch.cuda.amp.GradScaler()

    def load_checkpoint(self, path):
        checkpoint = torch.load(path)
        self.model.load_state_dict(checkpoint['model']) # type: ignore
        self.optimizer.load_state_dict(checkpoint['optimizer']) # type: ignore
        print('  Load model from', path)

    def save_checkpoint(self, path):
        state = {'model': self.model.state_dict(), 'optimizer': self.optimizer.state_dict()} # type: ignore
        torch.save(state, path)
        print('  Model is saved in', path)

    def train(self, data, epochs, w_cls, w_dae, w_gae):
        self.model.train() # type: ignore
        start_time = time.time()
        for epoch in range(1, epochs + 1):
            data = data.to(self.device, non_blocking=True)
            inputs, targets = data.x, data.y
            edge_index = data.edge_index
            edge_weight = data.edge_weight
            with torch.cuda.amp.autocast():
                outputs, dae_loss, gae_loss = self.model(inputs, edge_index, edge_weight) # type: ignore
                loss = w_cls * self.criterion(outputs, targets) + w_dae * dae_loss + w_gae * gae_loss # type: ignore
            train_loss = loss.item()
            self.optimizer.zero_grad() # type: ignore
            self.scaler.scale(loss).backward() # type: ignore
            self.scaler.step(self.optimizer) # type: ignore
            self.scaler.update() # type: ignore
            total = targets.size(0)
            predictions = outputs.argmax(1)
            correct = predictions.eq(targets.argmax(1)).sum().item()
            self.scheduler.step() # type: ignore
            process_time = time.time() - start_time
            accuracy = correct / total * 100.0
            print('  [Epoch %3d] Loss: %.5f, Time: %.2f s, Psuedo-Acc: %.2f%%'
                  % (epoch, train_loss, process_time, accuracy))

    def valid(self, data):
        self.model.eval() # type: ignore
        with torch.no_grad():
            data = data.to(self.device)
            inputs = data.x
            edge_index = data.edge_index
            edge_weight = data.edge_weight
            outputs, _, _ = self.model(inputs, edge_index, edge_weight) # type: ignore
            predictions = outputs.argmax(1)
        predictions = predictions.detach().cpu().numpy()
        return predictions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/2/28 16:09
# @Author  : zhoutao
# @File    : setup.py.py
# @Software: VScode
# @Email   : zhotoa@foxmail.com

import setuptools
from wheel.bdist_wheel import bdist_wheel

__version__ = "1.0.7"


class BDistWheel(bdist_wheel):
    def get_tag(self):
        return (self.python_tag, "none", "any")


cmdclass = {
    "bdist_wheel": BDistWheel,
}

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


# requirements = open("requirements.txt").readline()
install_requires = [
    "anndata>=0.8.0",
    "matplotlib>=3.5.1",
    "matplotlib-inline>=0.1.3",
    "networkx>=2.7.1",
    "numpy>=1.21.5",
    "pandas>=1.4.2",
    "scanpy>=1.9.1",
    "scikit-learn>=1.0.2",
    "scipy>=1.8.0",
    "seaborn>=0.11.2",
    "torch>=1.11.0",
    "torch-geometric>=2.0.2",
    "xgboost>=2.0.0"
    
]
setuptools.setup(
    name = "stTransfer",
    version = "1.0.10",
    description = "Transfer learning for spatial transcriptomics data and single-cell RNA-seq data.",
    author = "zhoutao",
    author_email = "zhotoa@foxmail.com",
    url = "https://github.com/zepoch/stTransfer.git",
    python_requires=">=3.8",
    packages=setuptools.find_packages(),
    # install_requires=requirements,
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    cmdclass=cmdclass,
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 25/2/24 21:54 PM
# @Author  : zhoutao3
# @File    : *.py
# @Email   : zhotoa@foxmail.com

from pyexpat import model
import scipy.sparse as sp
import numpy as np
import xgboost as xgb # type: ignore
import anndata as ad
import os.path as osp
import pandas as pd
import scanpy as sc # type: ignore
import numpy as np
import scipy.sparse as sp
import random
import torch
import torch_geometric # type: ignore
from matplotlib import pyplot as plt

from .cell_type_ann_model import SpatialModelTrainer # type: ignore

from typing import Dict, Optional, Tuple, Sequence,List

def adata_precess(adata: ad.AnnData,):
    '''
    adata: ad.AnnData, scRNA-seq data
    '''
    data = adata.X.toarray() if sp.issparse(adata.X) else adata.X # Fix: Use .toarray() to convert sparse matrix to dense matrix
    
    norm_factor = np.linalg.norm(data, axis=1, keepdims=True) # Fix: Use np.linalg.norm to calculate the norm of each row
    norm_factor[norm_factor == 0] = 1
    data = data / norm_factor
    
    return data

def xgboost_train(X: np.ndarray, 
                  y: np.ndarray, 
                  save_path: str,
                  n_fold: int = 10,
                  gpu: Optional[str] = None,):
    '''
    X: np.ndarray, scRNA-seq data
    y: np.ndarray, cell type annotation
    save_path: str, model save path
    n_fold: int, number of folds for xgboost training
    '''
    if gpu is not None and torch.cuda.is_available():
        device = "cuda:{}".format(gpu)
    else:
        device = "cpu"
        
    from sklearn.model_selection import train_test_split
    from sklearn.datasets import load_breast_cancer
    from sklearn.model_selection import KFold
    import pickle
    
    dic = {list(set(y))[i]:i for i in range(len(set(y)))}
    y = np.array([dic[i] for i in y])
    reverse_dic = {v:k for k,v in dic.items()}
    # X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    print('########--- model init ---##########')
    model = xgb.XGBClassifier(objective='multi: softmax', n_estimators=100, seed=42, )
    kf = KFold(n_splits=n_fold, random_state=42, shuffle=True)
    # 用于存储每折的分数
    scores = []
    print('########--- model train ---##########')
    count = 0
    for train_index, test_index in kf.split(X):
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]

        # 在训练集上训练模型
        model.fit(X_train, y_train, verbose=2)

        # 在测试集上评估模型
        score = model.score(X_test, y_test)
        print(f"Fold {count+1}: {score}")
        
        scores.append(score)
        count += 1

    # 打印平均分数
    print(f"Average score: {np.mean(scores)}")
    
    with open(osp.join(save_path,'xgboost_model.pkl'), 'wb') as f:
        pickle.dump((model, reverse_dic), f)
    return reverse_dic
    
def xgboost_fit(X: np.ndarray,
                save_path: str,
                dic: Dict,
                model_name: str = 'xgboost_model.pkl',):
    import pickle
    if model_name:
        with open(osp.join(save_path, model_name), 'rb') as f:
            model, dic = pickle.load(f)
    else:
        with open(osp.join(save_path, 'xgboost_model.pkl'), 'rb') as f:
            model, dic = pickle.load(f)
    
    psuedo_label = model.predict_proba(X)
    psuedo_class = [dic[i] for i in psuedo_label.argmax(1)]
    return psuedo_label, psuedo_class

import os
 
def mkdir(path):
 
	folder = os.path.exists(path)
	if not folder:                   
		os.makedirs(path)            
		print("---  new folder...  ---")
 
	else:
		print("---  There is this folder!  ---")
		
def distribution_fine_tune(X: np.ndarray, 
                           cell_coo: np.ndarray,
                           psuedo_label: np.ndarray,
                           psuedo_classes: Dict,
                           pca_dim: int = 500, 
                           k_graph: int = 30, 
                           edge_weight: bool = True, 
                           epochs: int = 200, 
                           w_cls: int = 50, 
                           w_dae: int = 1, 
                           w_gae: int = 1,
                           gpu: str = "0", 
                           save_path: str = "./output"):
    """
    :param adata:
    :param pca_dim: PCA dims, default=200
    :param k_graph: neighbors number, default=30
    :param edge_weight: Add edge weight to the graph model, default=True
    :param epochs: GCN training epochs, default=200
    :param w_cls: class num weight, default=20
    :param w_dae: dnn weight
    :param w_gae: gcn weight
    :param gpu: gpu number or None for cpu
    :param save_path: results save path
    :return:
    """
    if gpu is not None and torch.cuda.is_available():
        device = torch.device("cuda:{}".format(gpu))
    else:
        device = torch.device("cpu")

    print("========> Model Training...")
    
    gene_mat = torch.Tensor(X)
    if pca_dim:
        u, s, v = torch.pca_lowrank(gene_mat, pca_dim)
        gene_mat = torch.matmul(gene_mat, v)
    # u, s, v = torch.pca_lowrank(gene_mat, pca_dim)
    # gene_mat = torch.matmul(gene_mat, v)
    cell_coo = torch.Tensor(cell_coo) # type: ignore
    data = torch_geometric.data.Data(x=gene_mat, pos=cell_coo) # type: ignore
    data = torch_geometric.transforms.KNNGraph(k=k_graph, loop=True)(data) # type: ignore
    data.y = torch.Tensor(psuedo_label)

    # Make distances as edge weights.
    if edge_weight:
        data = torch_geometric.transforms.Distance()(data) # type: ignore
        data.edge_weight = 1 - data.edge_attr[:, 0]
    else:
        data.edge_weight = torch.ones(data.edge_index.size(1))

    # Train self-supervision model.
    input_dim = data.num_features
    num_classes = len(psuedo_classes)
    trainer = SpatialModelTrainer(input_dim, num_classes, device=device)
    trainer.train(data, epochs, w_cls, w_dae, w_gae)
    trainer.save_checkpoint(osp.join(save_path, "graph_finetune.bgi"))

    # Inference.
    print('\n==> Inferencing...')
    predictions = trainer.valid(data)
    celltype_pred = pd.Categorical([psuedo_classes[i] for i in predictions])
    pd.DataFrame([psuedo_classes[i] for i in predictions]).to_csv(osp.join(save_path, 'celltype_pred.csv'))
    # celltype_pred.to_csv(osp.join(save_path, 'celltype_pred.csv'))
    return celltype_pred

def sc_model_train_test(sc_adata: ad.AnnData,
                        st_adata: ad.AnnData,
                        sc_ann_key: str,
                        save_path: str,
                        marker_genes: Optional[List[str]] = None,
                        st_adata_spatial_key: str = 'spatial',
                        finetune_epochs: int = 50,
                        finutune_pca_dim: int = 500,):
    mkdir(save_path)
    print('########--- pre process ---##########')
    sc_adata_var_names = sc_adata.var_names.tolist()
    st_adata_var_names = st_adata.var_names.tolist()
    common_genes = list(set(sc_adata_var_names) & set(st_adata_var_names))
    assert len(common_genes) > 0, 'No common genes between scRNA-seq and spatial transcriptomics data'
    sc_adata = sc_adata[:, sc_adata.var_names.isin(common_genes)]
    st_adata = st_adata[:, st_adata.var_names.isin(common_genes)]
    # re-index gene order
    sc_adata = sc_adata[:, np.array(common_genes).tolist()]
    st_adata = st_adata[:, np.array(common_genes).tolist()]
    
    sc_X = adata_precess(sc_adata)
    ST_X = adata_precess(st_adata)
    sc_y = np.array(sc_adata.obs[sc_ann_key].to_list())
    print('########--- start trian ---##########')
    reverse_dic = xgboost_train(sc_X, sc_y, save_path)  # Fix: Pass the correct arguments to xgboost_train
    
    psuedo_label, psuedo_class = xgboost_fit(ST_X, dic = reverse_dic, save_path = save_path)  # Fix: Pass the correct arguments to xgboost_fit
    pd.DataFrame(psuedo_label).to_csv(osp.join(save_path, 'psuedo_label.csv'))
    pd.DataFrame(psuedo_class).to_csv(osp.join(save_path, 'psuedo_class.csv'))
    cell_coo = st_adata.obsm[st_adata_spatial_key]
    distribution_fine_tune(ST_X, 
                           cell_coo = cell_coo, 
                           gpu = None,
                           psuedo_label = psuedo_label,  # Add the missing argument "psuedo_label"
                           psuedo_classes = reverse_dic,  # Add the missing argument "psuedo_classes"
                           save_path=save_path,
                           epochs = finetune_epochs,
                           pca_dim = finutune_pca_dim)